# -*- coding: utf-8 -*-
import sys
import re
import unittest
from StringIO import StringIO
from skoolkittest import SkoolKitTestCase
from skoolkit.skoolasm import SkoolWriter
from skoolkit.skoolparser import SkoolParser, CASE_LOWER

TEST_SKOOL = 'test-asm.skool'

TABLE_SRC = []
TABLE_TEXT = []

# Headers, colspans 1 and 2, rowspans 1 and 2, transparent cell
TABLE_SRC.append("""(data)
{ =h Col1 | =h Col2 | =h,c2 Cols3+4 }
{ =r2 X   | Y       | Za  | Zb }
{           Y2      | Za2 | =t }""")

TABLE_TEXT.append("""
+------+------+----------+
| Col1 | Col2 | Cols3+4  |
+------+------+-----+----+
| X    | Y    | Za  | Zb |
|      | Y2   | Za2 |    |
+------+------+-----+----+
""")

# Cell with rowspan > 1 in the middle column
TABLE_SRC.append("""
{ A1 cell | =r2 This is a cell with rowspan=2 | C1 cell }
{ A2 cell | C2 cell }""")

TABLE_TEXT.append("""
+---------+-------------------------------+---------+
| A1 cell | This is a cell with rowspan=2 | C1 cell |
| A2 cell |                               | C2 cell |
+---------+-------------------------------+---------+
""")

# Cell with rowspan > 1 in the rightmost column
TABLE_SRC.append("""(,,,:w)
{ A1 cell | B1 cell | =r2 This is a cell with rowspan=2 }
{ A2 cell | B2 cell }""")

TABLE_TEXT.append("""
+---------+---------+-------------------------------+
| A1 cell | B1 cell | This is a cell with rowspan=2 |
| A2 cell | B2 cell |                               |
+---------+---------+-------------------------------+
""")

class SkoolWriterTest(SkoolKitTestCase):
    def _get_writer(self, crlf=False, tab=False, lower=False, instr_width=23, warn=False):
        case = CASE_LOWER if lower else None
        skool_parser = SkoolParser(TEST_SKOOL, case=case, asm_mode=1)
        return SkoolWriter(skool_parser, crlf, tab, lower, instr_width, warn)

    def _get_asm(self, crlf=False, tab=False, lower=False, instr_width=23, warn=False):
        writer = self._get_writer(crlf, tab, lower, instr_width, warn)
        out = StringIO()
        err = StringIO() if warn else None
        writer.write(out, err)
        asm = out.getvalue().split('\n')[:-1]
        if warn:
            return asm, err.getvalue().split('\n')[:-1]
        return asm

    def _test_unsupported_macro(self, text):
        search = re.search('#[A-Z]+', text)
        macro = search.group()
        writer = self._get_writer()
        writer.err = StringIO()
        self.assertRaises(SystemExit, writer.strip_macros, text)
        self.assertEqual('ERROR: Found unsupported macro: %s\n' % macro, writer.err.getvalue())

    def _test_reference_macro(self, macro, def_link_text):
        writer = self._get_writer()
        for link_text in ('',  '(testing)', '(testing (nested) parentheses)'):
            for anchor in ('', '#name'):
                output = writer.strip_macros('#%s%s%s' % (macro, anchor, link_text))
                self.assertEqual(output, link_text[1:-1] or def_link_text)

    def test_macro_bug(self):
        self._test_reference_macro('BUG', 'bug')

    def test_macro_c(self):
        writer = self._get_writer()
        output = writer.strip_macros('#C')
        self.assertEqual(output, '(c)')

    def test_macro_call(self):
        self._test_unsupported_macro('#CALL:method()')

    def test_macro_d(self):
        writer = self._get_writer()

        output = writer.strip_macros('#D24581')
        self.assertEqual(output, 'Another routine')

        output = writer.strip_macros('#D$6010')
        self.assertEqual(output, 'Yet another routine still')

    def test_macro_erefs(self):
        self._test_unsupported_macro('#EREFS24576')

    def test_macro_fact(self):
        self._test_reference_macro('FACT', 'fact')

    def test_macro_font(self):
        self._test_unsupported_macro('#FONT55584,96,,,1')

    def test_macro_link(self):
        writer = self._get_writer()
        link_text = 'Testing'
        output = writer.strip_macros('#LINK:PageID#anchor(%s)' % link_text)
        self.assertEqual(output, link_text)

    def test_macro_name(self):
        writer = self._get_writer()
        output = writer.strip_macros('#NAMEanchor')
        self.assertEqual(output, '')

    def test_macro_p(self):
        writer = self._get_writer()
        output = writer.strip_macros('#P')
        self.assertEqual(output, '')

    def test_macro_poke(self):
        self._test_reference_macro('POKE', 'poke')

    def test_macro_pokes(self):
        self._test_unsupported_macro('#POKES24576,1,10,2')

    def test_macro_pops(self):
        self._test_unsupported_macro('#POPS')

    def test_macro_pushs(self):
        self._test_unsupported_macro('#PUSHSname')

    def test_macro_r(self):
        writer = self._get_writer()

        output = writer.strip_macros('#R24576')
        self.assertEqual(output, 'DOSTUFF')

        link_text = 'Testing1'
        output = writer.strip_macros('#R24576(%s)' % link_text)
        self.assertEqual(output, link_text)

        output = writer.strip_macros('#R24579')
        self.assertEqual(output, '6003')

        output = writer.strip_macros('#R24580')
        self.assertEqual(output, '6004')

        output = writer.strip_macros('#R32768@other')
        self.assertEqual(output, '32768')

        link_text = 'Testing2'
        output = writer.strip_macros('#R32768@other#testing(%s)' % link_text)
        self.assertEqual(output, link_text)

    def test_macro_refs(self):
        self._test_unsupported_macro('#REFS24576')

    def test_macro_reg(self):
        writer = self._get_writer()
        for reg in ("a", "b", "c", "d", "e", "h", "l", "i", "r", "ixl", "ixh", "iyl", "iyh", "b'", "c'", "d'", "e'", "h'", "l'", "bc", "de", "hl", "sp", "ix", "iy", "bc'", "de'", "hl'"):
            output = writer.strip_macros('#REG%s' % reg)
            self.assertEqual(output, reg.upper())

    def test_macro_scr(self):
        self._test_unsupported_macro('#SCR2(fname)')

    def test_macro_space(self):
        writer = self._get_writer()

        output = writer.strip_macros('"#SPACE"')
        self.assertEqual(output, '" "')

        output = writer.strip_macros('"#SPACE5"')
        self.assertEqual(output, '"     "')

    def test_macro_table(self):
        writer = self._get_writer()

        for src, text in zip(TABLE_SRC, TABLE_TEXT):
            table = '#TABLE%s\nTABLE#' % src

            output = writer.strip_macros(table)
            self.assertEqual(output, table)

            output = '\n'.join(writer.format(table, 79))
            self.assertEqual(output, text.strip())

    def test_macro_udg(self):
        self._test_unsupported_macro('#UDG39144,6(safe_key)')

    def test_macro_udgarray(self):
        self._test_unsupported_macro('#UDGARRAY8,,,256;33008-33023(bubble)')

    def test_macro_udgtable(self):
        writer = self._get_writer()
        udgtable = '#UDGTABLE%s\nUDGTABLE#' % TABLE_SRC

        output = writer.strip_macros(udgtable)
        self.assertEqual(output, udgtable)

        output = writer.format(udgtable, 79)
        self.assertEqual(output, [])

    def test_option_crlf(self):
        asm = self._get_asm(crlf=True)
        self.assertTrue(all(line.endswith('\r') for line in asm))

    def test_option_tab(self):
        asm = self._get_asm(tab=True)
        for line in asm:
            if line and line[0].isspace():
                self.assertEqual(line[0], '\t')

    def test_option_lower(self):
        asm = self._get_asm(lower=True)

        # Test that the ORG directive is in lower case
        self.assertEqual(asm[0], '  org 24576')

        # Test that labels are unaffected
        self.assertEqual(asm[8], 'DOSTUFF:')

        # Test that #REG macro arguments are in lower case
        self.assertEqual(asm[14], '  defb 123                ; a=0')

        # Test that #R macro arguments that resolve to a label are unaffected
        self.assertEqual(asm[15], '  defb 246                ; DOSTUFF')

    def test_option_instr_width(self):
        for width in (5, 10, 15, 20, 25, 30):
            asm = self._get_asm(instr_width=width)
            self.assertEqual(asm[14], '  %s ; A=0' % 'DEFB 123'.ljust(width))

    def test_option_warn(self):
        asm, err = self._get_asm(warn=True)

        # Test warning about #R macro with address that has no label
        self.assertTrue('WARNING: Could not convert address $6003 to label' in err)

if __name__ == '__main__':
    unittest.main()
